# Warning: Your eyes may never recover from the mess this code is.

import sys
import random
import pygame

import data
import level
import const
import util


class GameWindow(object):

    def __init__(self):
        #icon = pygame.image.load(data.filepath(const.IICON))
        #pygame.display.set_icon(icon)
        pygame.display.set_caption(const.TITLE)

        self.screen = pygame.display.set_mode((const.SIZE_F*const.WIDTH,
                                               const.SIZE_F*const.HEIGHT))

        self.game()

#    def menu(self):
#        Menu(self, self.screen).loop()

    def game(self):
        Game(self, self.screen).loop()


class Game(object):

    def __init__(self, window, screen):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.WIDTH, const.HEIGHT))

        self.clock = pygame.time.Clock()

        self.paused = False

        self.bg = pygame.image.load(data.filepath('bg.png')).convert()
        self.bg_trees = MovingBackground(0)
        self.bg_mountains = MovingBackground(1)

        self.bullet_mgr = BulletManager()
        self.feather_mgr = FeatherManager()
        self.lives = Lives(const.LIVES)
        self.character = Character(self.bullet_mgr, self.feather_mgr,
                                   self.lives)
        self.up_marker = UpMarker(self.character)
        self.character_con = pygame.sprite.RenderUpdates(self.character,
                                                         self.up_marker)
        self.enemy_bullet_mgr = EnemyBulletManager()
        self.enemy_mgr = EnemyManager(self.enemy_bullet_mgr)

        self.collision_mgr = CollisionManager(self.character,
                                              self.feather_mgr,
                                              self.bullet_mgr,
                                              self.enemy_mgr,
                                              self.enemy_bullet_mgr)

        self.level = Level(0, self.enemy_mgr, self)

        self.message = Message()
        self.message_con = pygame.sprite.RenderUpdates(self.message)

        self.restarting = False

        pygame.mixer.init()
        pygame.mixer.music.load(data.filepath('8bp088-02-nullsleep-supernova_kiss.ogg'))

    def loop(self):
        self.screen.blit(self.bg, (0, 0))
        pygame.transform.scale(self.screen,
                               (const.SIZE_F*const.WIDTH,
                                const.SIZE_F*const.HEIGHT),
                               self.real_screen)
        pygame.display.update()

        pygame.mixer.music.play()

        self.done = False

        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)

            if not self.paused:
                self.bg_mountains.clear(self.screen, self.bg)
                self.bg_trees.clear(self.screen, self.bg)
                self.enemy_mgr.clear(self.screen, self.bg)
                self.feather_mgr.clear(self.screen, self.bg)
                self.character_con.clear(self.screen, self.bg)
                self.bullet_mgr.clear(self.screen, self.bg)
                self.enemy_bullet_mgr.clear(self.screen, self.bg)
                self.lives.clear(self.screen, self.bg)
            self.message_con.clear(self.screen, self.bg)

            dt = self.clock.tick(const.FPS) / 1000.0
            if not self.paused:
                self.bg_mountains.update(dt)
                self.bg_trees.update(dt)
                self.level.update(dt)
                self.character.update(dt)
                self.up_marker.update(dt)
                self.feather_mgr.update(dt)
                self.bullet_mgr.update(dt)
                self.enemy_mgr.update(dt)
                self.enemy_bullet_mgr.update(dt)
                self.collision_mgr.update(dt)
                self.message.update(dt)

            update_rects = []
            if not self.paused:
                update_rects += self.bg_mountains.draw(self.screen)
                update_rects += self.bg_trees.draw(self.screen)
                update_rects += self.enemy_mgr.draw(self.screen)
                update_rects += self.feather_mgr.draw(self.screen)
                update_rects += self.character_con.draw(self.screen)
                update_rects += self.bullet_mgr.draw(self.screen)
                update_rects += self.enemy_bullet_mgr.draw(self.screen)
                update_rects += self.lives.draw(self.screen)
            update_rects += self.message_con.draw(self.screen)

            pygame.transform.scale(self.screen,
                                   (const.SIZE_F*const.WIDTH,
                                    const.SIZE_F*const.HEIGHT),
                                   self.real_screen)
            #screen2 = pygame.transform.scale2x(self.screen)
            #pygame.transform.scale2x(screen2, self.real_screen)

            update_rects_scaled = []
            for rect in update_rects:
                update_rects_scaled.append(pygame.Rect(
                    rect.left * const.SIZE_F,
                    rect.top * const.SIZE_F,
                    rect.width * const.SIZE_F,
                    rect.height * const.SIZE_F))
            pygame.display.update(update_rects_scaled)
            #pygame.display.update() # 200x120 is so small...

            if self.character.lost:
                self.restart()

            if self.level.done and len(self.enemy_mgr.sprites()) == 0:
                self.win()
                
        if self.restarting: self.window.game()

    def pause(self):
        if self.paused:
            self.paused = False
            self.message.clear_message()
            pygame.display.update()
        else:
            self.paused = True
            self.message.write('Paused')
            pygame.display.update()

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            sys.exit() # TODO something else
        #elif event.key == pygame.K_p:
        #    self.pause()
        elif event.key == pygame.K_r:
            self.restart()        
        elif event.key == pygame.K_x or event.key == pygame.K_SPACE:
            self.character.start_action()
        elif event.key == pygame.K_UP or event.key == pygame.K_w:
            self.character.go_up()
        elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
            self.character.go_down()
        elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
            self.character.go_left()
        elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
            self.character.go_right()

    def on_keyup(self, event):
        if event.key == pygame.K_x or event.key == pygame.K_SPACE:
            self.character.stop_action()
        elif event.key == pygame.K_UP or event.key == pygame.K_w:
            self.character.stop_up()
        elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
            self.character.stop_down()
        elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
            self.character.stop_left()
        elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
            self.character.stop_right()

    def feather_level_up(self):
        self.feather_mgr.level_up()
        if self.feather_mgr.max_feather == 1:
            self.message.write('Orange feather joins the battle!')
        elif self.feather_mgr.max_feather == 2:
            self.message.write('Blue feather joins the battle!')
        else:
            self.message.write('Green feather joins the battle!')
        self.message.timer = 3

    def speed_level_up(self):
        self.feather_mgr.level_up_rs()
        self.message.write('Faster!')
        self.message.timer = 3

    def attack_level_up(self):
        self.character.level_up()
        if self.character.max_level == 1:
            self.message.write('Same-color combo 2x available!')
        else:
            self.message.write('Same-color combo 3x available!')
        self.message.timer = 3

    def restart(self):
        pygame.mixer.stop()
        self.done = True
        self.restarting = True

    def win(self):
        self.paused = True
        self.message.write('Win!')
        pygame.display.update()


class Character(pygame.sprite.Sprite):

    def __init__(self, bullet_mgr, feather_mgr, lives):
        pygame.sprite.Sprite.__init__(self)

        self.bullet_mgr = bullet_mgr

        self.pos = [0, 0]
        self.vel = [0, 0]
        self.acc = [0, 0]

        self.image_sit = pygame.image.load(data.filepath('character.png')
                                       ).convert()
        self.image_sit.set_colorkey((255, 0, 255))
        self.image_up = pygame.image.load(data.filepath('character-up.png')
                                       ).convert()
        self.image_up.set_colorkey((255, 0, 255))                            
        self.image_down = pygame.image.load(data.filepath(
            'character-down.png')).convert()
        self.image_down.set_colorkey((255, 0, 255))
        self.image_null = pygame.surface.Surface((0, 0))
        self.image_null.fill((0, 0, 0))
        self.image_null.set_colorkey((0, 0, 0))
        self.prev_image = None
        
        self.image = self.image_sit
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.on_air = True
        self.jumping = False
        self.jump_timer = 0
        self.going_left = False
        self.going_right = False
        self.going_up = False
        self.going_down = False

        self.action = False
        self.lost = False

        self.feather = None

        self.lives = lives

        self.invul_timer = -100

        self.max_level = 0

        self.reset_bullets()

        self.feather_mgr = feather_mgr
        self.respawn()

    def start_action(self):
        self.action = True
        if not self.on_air:
            self.jump()

    def stop_action(self):
        self.action = False
        self.stop_jump()

    def jump(self):
        self.on_air = True
        self.let_feather_go()
        self.jumping = True
        self.jump_timer = const.JUMP_TIMER
        self.vel[1] = -const.JUMP_VEL
        self.image = self.image_up

    def fall(self):
        self.on_air = True
        self.let_feather_go()

    def let_feather_go(self):
        if self.feather:
            self.feather.go()
            self.feather = None

    def stop_jump(self):
        self.jumping = False
        self.jump_timer = 0

    def land(self, feather):
        self.on_air = False
        self.feather = feather
        self.image = self.image_sit
        if not self.feather.activated:
            self.bullet_timers[self.feather.item_num] = const.BULLET_TIME
            if self.bullet_levels[self.feather.item_num] < self.max_level:
                self.bullet_levels[self.feather.item_num] += 1

    def level_up(self):
        if self.max_level < len(const.TYPES) - 1:
            self.max_level += 1

    def go_left(self): self.going_left = True
    def stop_left(self): self.going_left = False
    def go_right(self): self.going_right = True
    def stop_right(self): self.going_right = False
    def go_up(self): self.going_up = True
    def stop_up(self): self.going_up = False
    def go_down(self): self.going_down = True
    def stop_down(self): self.going_down = False

    def fire(self, item, level):
        self.bullet_mgr.generate_bullets(self.pos, item, level)

    def respawn(self, invul=False):
        self.pos = [40, const.HEIGHT/2]
        self.vel = [0, 0]
        self.feather_mgr.generate_feather_specific((40, const.HEIGHT/2 + 3))
        self.reset_bullets()
        if invul:
            self.invul_timer = const.INVUL_TIMER

    def reset_bullets(self):
        self.bullet_timers = [0, 0, 0, 0]
        self.bullet_rofs = [0, 0, 0, 0]
        self.bullet_levels = [-1, -1, -1, -1]

    def die(self):
        if self.invul_timer > 0: return
        self.fall()
        self.respawn(True)
        if not self.lives.lose():
            self.lose()

    def lose(self):
        self.lost = True

    def update_gravity(self, dt):
        if self.on_air:
            self.acc[1] = const.GRAVITY
        else:
            self.acc[0] = 0
            self.acc[1] = 0

    def update_jump(self, dt):
        if self.jumping:
            self.vel[1] = -const.JUMP_VEL
            self.jump_timer -= dt
            if self.jump_timer <= 0:
                self.stop_jump()

    def update_move(self, dt):
        if self.on_air:
            if (self.image != self.image_down and self.vel[1] > 0
                and self.image != self.image_null):
                self.image = self.image_down
            if self.going_left: self.acc[0] = -const.MOVE_ACC
            elif self.going_right: self.acc[0] = const.MOVE_ACC
            else:
                self.acc[0] = 0
                self.vel[0] *= 0.8
                if abs(self.vel[0]) < 1: self.vel[0] = 0
        else:
            if self.going_left: self.vel[0] = -const.MOVE_F_VEL
            elif self.going_right: self.vel[0] = const.MOVE_F_VEL
            else: self.vel[0] = 0
            if self.going_up: self.vel[1] = -const.MOVE_F_VEL
            elif self.going_down: self.vel[1] = const.MOVE_F_VEL
            else: self.vel[1] = 0

    def update_fire(self, dt):
        for i in range(len(self.bullet_timers)):
            if self.bullet_timers[i] > 0:
                self.bullet_timers[i] -= dt
                self.bullet_rofs[i] -= dt
                level = self.bullet_levels[i]
                if self.bullet_rofs[i] <= 0:
                    self.bullet_rofs[i] = const.BULLET_TYPES[i]['rof'][level]
                    self.fire(i, level)
            else:
                self.bullet_timers[i] = 0
                self.bullet_rofs[i] = 0
                self.bullet_levels[i] = -1

    def update_invul(self, dt):
        if self.invul_timer > 0:
            self.invul_timer -= dt
            if self.image == self.image_null:
                self.image = self.prev_image
                self.prev_image = None
            else:
                self.prev_image = self.image
                self.image = self.image_null
            if self.invul_timer <= 0:
                self.invul_timer = -100
                if self.image == self.image_null:
                    self.image = self.prev_image
                    self.prev_image = None


    def update(self, dt):
        self.update_gravity(dt)
        self.update_jump(dt)
        self.update_move(dt)
        self.update_fire(dt)
        self.update_invul(dt)

        self.vel[0] += self.acc[0] * dt
        self.vel[1] += self.acc[1] * dt
        if self.vel[0] > const.VEL_LIMIT_X: self.vel[0] = const.VEL_LIMIT_X
        if self.vel[0] < -const.VEL_LIMIT_X: self.vel[0] = -const.VEL_LIMIT_X
        if self.vel[1] > const.VEL_LIMIT_Y: self.vel[1] = const.VEL_LIMIT_Y
        #if self.vel[1] < -const.VEL_LIMIT_Y: self.vel[1] = -const.VEL_LIMIT_Y
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt

        # bound
        if self.pos[0] < self.rect.width/2:
            self.pos[0] = self.rect.width/2
        if self.pos[0] > const.WIDTH - self.rect.width/2:
            self.pos[0] = const.WIDTH - self.rect.width/2
        if not self.on_air:
            if self.pos[1] < self.rect.height/2:
                self.pos[1] = self.rect.height/2
            if self.pos[1] > const.HEIGHT - self.rect.height/2:
                self.pos[1] = const.HEIGHT - self.rect.height/2

        if self.pos[1] > const.HEIGHT + self.rect.height:
            self.die()
        
        self.rect.center = self.pos


class UpMarker(pygame.sprite.Sprite):

    def __init__(self, character):
        pygame.sprite.Sprite.__init__(self)
        
        self.pos = [-100, -100]

        self.image = pygame.image.load(data.filepath('arrow-up.png')
                                       ).convert()
        self.image.set_colorkey((255, 0, 255))
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.character = character

    def update(self, dt):
        if self.character.pos[1] < -self.character.rect.height:
            self.pos[0] = self.character.pos[0]
            self.pos[1] = self.rect.height
        else:
            self.pos = [-100, -100]
        self.rect.center = self.pos


class FeatherManager(pygame.sprite.RenderUpdates):

    def __init__(self):
        pygame.sprite.RenderUpdates.__init__(self)

        self.max_feather = 0
        self.rs_level = 0
        self.feather_respawn = const.FEATHER_RESPAWN[self.rs_level]
        self.timer = self.feather_respawn

    def level_up(self):
        if self.max_feather < len(const.TYPES) - 1:
            self.max_feather += 1

    def level_up_rs(self):
        if self.rs_level < len(const.FEATHER_VEL) - 1:
            self.rs_level += 1
            self.feather_respawn = const.FEATHER_RESPAWN[self.rs_level]

    def update(self, dt):
        self.timer -= dt
        if self.timer <= 0:
            self.timer = self.feather_respawn
            self.generate_feather()
        for feather in iter(self):
            feather.update(dt)
            if feather.done: self.remove(feather)

    def generate_feather(self):
        height = random.randint(const.FEATHER_HEIGHT, const.HEIGHT -
                                const.FEATHER_HEIGHT)
        feather = Feather((const.WIDTH + const.FEATHER_WIDTH / 2, height),
                          const.FEATHER_VEL[self.rs_level],
                          random.randint(0, self.max_feather))
        self.add(feather)

    def generate_feather_specific(self, pos):
        feather = Feather(pos, const.FEATHER_VEL[self.rs_level],
                          random.randint(0, self.max_feather))
        self.add(feather)
        

class Feather(pygame.sprite.Sprite):

    def __init__(self, pos, vel, item):
        pygame.sprite.Sprite.__init__(self)

        self.pos = list(pos)
        self.vel = [0, 0]
        self.item_num = item
        self.item = const.TYPES[item]

        self.image2 = pygame.image.load(data.filepath(self.item['name'] +
                                                     '2.png')).convert()
        self.image2.set_colorkey((255, 0, 255))
        self.image3 = pygame.image.load(data.filepath(self.item['name'] +
                                                     '3.png')).convert()
        self.image3.set_colorkey((255, 0, 255))
        self.image4 = pygame.image.load(data.filepath(self.item['name'] +
                                                     '4.png')).convert()
        self.image4.set_colorkey((255, 0, 255))

        self.image = pygame.image.load(data.filepath(self.item['name'] +
                                                     '.png')).convert()
        self.image.set_colorkey((255, 0, 255))
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.vel[0] = -vel
        
        self.staying = False
        self.done = False
        self.faded = False

        self.character = None

        self.fade_timer = 8

        self.activated = False

    def go(self):
        self.staying = False
        self.character = None

    def stay(self, character):
        self.staying = True
        self.character = character
        self.activated = True

    def update(self, dt):
        if not self.staying:
            self.pos[0] += self.vel[0] * dt
            if self.pos[0] + const.FEATHER_WIDTH / 2 < 0:
                self.done = True
        else:
            self.pos[0] = self.character.pos[0]
            self.pos[1] = self.character.pos[1] + self.character.rect.height/2
            self.fade_timer -= dt
            if self.fade_timer <= -0.5:
                self.done = True
            if self.fade_timer <= 0:
                self.faded = True
                self.image = self.image4
                self.character.fall()
                self.go()
            elif self.fade_timer <= 2:
                self.image = self.image3
            elif self.fade_timer <= 4:
                self.image = self.image2
        self.rect.center = self.pos


class CollisionManager(object):

    def __init__(self, character, feather_mgr, bullet_mgr, enemy_mgr,
                 enemy_bullet_mgr):
        self.character = character
        self.feather_mgr = feather_mgr
        self.bullet_mgr = bullet_mgr
        self.enemy_mgr = enemy_mgr
        self.enemy_bullet_mgr = enemy_bullet_mgr

    def update(self, dt):
        # TODO more precise collision: only feet of character with center of feather
        col_list = pygame.sprite.spritecollide(self.character,
                                               self.feather_mgr, False,
                                               pygame.sprite.collide_rect)
        if (len(col_list) > 0 and self.character.on_air and
            self.character.vel[1] >= 0):
            for feather in col_list:
                if not feather.faded:
                    self.character.land(feather)
                    feather.stay(self.character)
                    break

        for enemy in iter(self.enemy_mgr):
            col_list = pygame.sprite.spritecollide(enemy, self.bullet_mgr,
                                            False, pygame.sprite.collide_rect)
            if len(col_list) > 0:
                bullet = col_list[0]
                enemy.hurt(bullet)
                self.bullet_mgr.remove(bullet)

        col_list = pygame.sprite.spritecollide(self.character,
                                               self.enemy_mgr, False,
                                               pygame.sprite.collide_rect)
        
        if len(col_list) > 0:
            self.character.die()
        
        col_list = pygame.sprite.spritecollide(self.character,
                                               self.enemy_bullet_mgr, False,
                                               pygame.sprite.collide_rect)
        
        if len(col_list) > 0:
            self.character.die()           


class BulletManager(pygame.sprite.RenderUpdates):

    def __init__(self):
        pygame.sprite.RenderUpdates.__init__(self)

    def generate_bullet(self, pos, vel, item, level):
        bullet = Bullet(pos, vel, item, level)
        self.add(bullet)

    def generate_bullets(self, pos, item, level):
        if item == 3: # yeah, yeah...
            vel = const.BULLET_TYPES[item]['vel'][level][0]
            npos = [pos[0], const.HEIGHT]
            self.generate_bullet(npos, vel, item, level)
            vel = const.BULLET_TYPES[item]['vel'][level][1]
            npos = [pos[0], 0]
            self.generate_bullet(npos, vel, item, level)
            if level == 2:
                vel = const.BULLET_TYPES[item]['vel'][level][2]
                npos = [pos[0] + 10, const.HEIGHT]
                self.generate_bullet(npos, vel, item, level)
                vel = const.BULLET_TYPES[item]['vel'][level][3]
                npos = [pos[0] + 10, 0]
                self.generate_bullet(npos, vel, item, level)
                vel = const.BULLET_TYPES[item]['vel'][level][4]
                npos = [pos[0] - 10, const.HEIGHT]
                self.generate_bullet(npos, vel, item, level)
                vel = const.BULLET_TYPES[item]['vel'][level][5]
                npos = [pos[0] - 10, 0]
                self.generate_bullet(npos, vel, item, level)
        else:
            for vel in const.BULLET_TYPES[item]['vel'][level]:
                self.generate_bullet(pos, vel, item, level)
                                 

    def update(self, dt):
        for bullet in iter(self):
            bullet.update(dt)
            if bullet.done:
                self.remove(bullet)


class Bullet(pygame.sprite.Sprite):

    def __init__(self, pos, vel, item, level):
        pygame.sprite.Sprite.__init__(self)

        self.pos = list(pos)
        self.vel = list(vel)

        self.image = pygame.image.load(data.filepath(
            const.BULLET_TYPES[item]['fname'][level])).convert()
        self.image.set_colorkey((255, 0, 255))

        # a few bullet hacks
        if item == 1 and vel[1] > 0:
            self.image = pygame.transform.flip(self.image, False, True)

        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.done = False

        self.damage = const.BULLET_TYPES[item]['damage'][level]

    def update(self, dt):
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        self.rect.center = self.pos
        if util.out_of_bounds(self):
            self.done = True


class EnemyManager(pygame.sprite.RenderUpdates):

    def __init__(self, enemy_bullet_mgr):
        pygame.sprite.RenderUpdates.__init__(self)
        
        self.enemy_bullet_mgr = enemy_bullet_mgr

    def generate_enemy(self, item, path):
        if path:
            enemy = Enemy(item, path, self.enemy_bullet_mgr)
            self.add(enemy)

    def update(self, dt):
        for enemy in iter(self):
            enemy.update(dt)
            if enemy.done: self.remove(enemy)


class Enemy(pygame.sprite.Sprite):

    def __init__(self, item, path, enemy_bullet_mgr):
        pygame.sprite.Sprite.__init__(self)

        self.enemy_bullet_mgr = enemy_bullet_mgr

        self.full_path = list(path)
        self.full_path_rev = list(path)
        self.full_path_rev.reverse()
        
        self.path = list(path)
        self.init_path()
        self.vel = [0, 0]
        self.item = item

        self.orig_image = pygame.image.load(data.filepath(
            'enemy-' + const.ENEMY_TYPES[item]['name'] + '.png')).convert()
        self.orig_image.set_colorkey((255, 0, 255))
        self.orig_image_flipped = pygame.transform.flip(self.orig_image, True,
                                                        False)

        self.image = self.orig_image
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.done = False

        self.wait_timer = 0
        self.reversed = False

        self.health = const.ENEMY_TYPES[item]['health']

        self.back_count = 0

    def pop_path(self):
        self.path.pop(0)
        self.path.pop(0)
        if len(self.path) == 0:
            # self.done = True
            #if not self.reversed:
            #    self.path = list(self.full_path_rev)
            #    self.image = self.orig_image_flipped
            #    self.reversed = True
            #else:
            #    self.path = list(self.full_path)
            #    self.image = self.orig_image
            #    self.reversed = False
            if self.back_count >= const.TIMES_BACK: self.done = True
            else:
                self.path = list(self.full_path)
                self.image = self.orig_image
                self.init_path()
                self.back_count += 1

    def init_path(self):
        self.pos = [self.path[0][0], self.path[0][1]]
        self.wait_timer = self.path[0][2]
        self.path.pop(0)

    def hurt(self, bullet):
        self.health -= bullet.damage
        if self.health <= 0:
            self.done = True
            # TODO boom

    def shoot(self):
        shot = const.ENEMY_TYPES[self.item]['shot']
        if shot:
            for bul in shot:
                self.enemy_bullet_mgr.generate_bullet(self.pos, bul[1],
                                                      bul[0])

    def update(self, dt):
        if self.done: return

        # assumes format is right... ugh
        if len(self.path) > 1:
            if self.wait_timer > 0:
                self.wait_timer -= dt
                if self.wait_timer <= 0:
                    self.wait_timer = 0
            else:
                target = [self.path[1][0], self.path[1][1]]
                self.pos = util.go_towards(self.pos, target, self.path[0], dt)
                if self.pos == target:
                    if len(self.path[1]) > 3:
                        self.shoot()
                    self.wait_timer = self.path[1][2]
                    self.pop_path()
        self.rect.center = self.pos


class MovingBackground(pygame.sprite.RenderUpdates):

    def __init__(self, item):
        pygame.sprite.RenderUpdates.__init__(self)

        self.sprite1 = pygame.sprite.Sprite()
        self.sprite1.image = pygame.image.load(data.filepath(
            const.MOVING_BG[item]['fname'])).convert()
        self.sprite1.image.set_colorkey((255, 0, 255))
        self.sprite1.rect = self.sprite1.image.get_rect()
        self.sprite1.rect.bottom = const.HEIGHT
        self.sprite1.rect.left = 0
        self.add(self.sprite1)

        self.sprite2 = pygame.sprite.Sprite()
        self.sprite2.image = self.sprite1.image.copy()
        self.sprite2.image.set_colorkey((255, 0, 255))
        self.sprite2.rect = self.sprite2.image.get_rect()
        self.sprite2.rect.bottom = const.HEIGHT
        self.sprite2.rect.left = const.WIDTH
        self.add(self.sprite2)

        self.left1 = 0
        self.left2 = const.WIDTH

        self.speed = const.MOVING_BG[item]['speed']

    def update(self, dt):
        self.left1 -= int(self.speed)
        self.left2 -= int(self.speed)
        if self.left1 + const.WIDTH <= 0:
            self.left1 += const.WIDTH
            self.left2 += const.WIDTH
        self.sprite1.rect.left = self.left1
        self.sprite2.rect.left = self.left2


class Lives(pygame.sprite.OrderedUpdates):

    def __init__(self, count):
        pygame.sprite.OrderedUpdates.__init__(self)

        self.count = 0

        self.pos = [2, 2]

        self.image = pygame.image.load(data.filepath('life.png')).convert()

        for i in range(count):
            self.gain()

    def lose(self):
        if self.count > 1:
            self.count -= 1
            self.remove(self.sprites()[-1])
            return True
        return False

    def gain(self):
        life = pygame.sprite.Sprite()
        life.image = self.image.copy()
        life.image.set_colorkey((255, 0, 255))
        life.rect = life.image.get_rect()
        life.rect.topleft = [self.pos[0] + self.count * 4, self.pos[1]]
        self.count += 1
        self.add(life)


class Level(object):

    def __init__(self, number, enemy_mgr, game):
        self.number = number
        self.enemy_mgr = enemy_mgr
        self.data = list(level.LEVEL[self.number])
        self.timer = 0
        self.game = game
        self.done = False

    def update(self, dt):
        self.timer += dt
        while len(self.data) > 0 and self.timer >= self.data[0][0]:
            if self.data[0][1] == 'f': self.game.feather_level_up()
            elif self.data[0][1] == 's': self.game.speed_level_up()
            elif self.data[0][1] == 'l': self.game.attack_level_up()
            elif self.data[0][1] == 'w': self.done = True
            else:
                self.enemy_mgr.generate_enemy(self.data[0][1],
                                              self.data[0][2])
            self.data.pop(0)
            

class Message(pygame.sprite.Sprite):

    def __init__(self):
        pygame.sprite.Sprite.__init__(self)

        pygame.font.init()
        self.pos = [const.WIDTH / 2, const.HEIGHT / 10]
        self.font = pygame.font.Font(data.filepath('slkscr.ttf'), 8)

        self.timer = 0
        
        self.clear_message()

    def clear_message(self):
        self.write('')

    def write(self, message):
        self.message = message
        self.image = self.font.render(self.message, False, (0, 0, 0))
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

    def update(self, dt):
        if self.timer > 0:
            self.timer -= dt
            if self.timer <= 0:
                self.timer = 0
                self.clear_message()


class EnemyBulletManager(pygame.sprite.RenderUpdates):

    def __init__(self):
        pygame.sprite.RenderUpdates.__init__(self)

    def generate_bullet(self, pos, vel, name):
        bullet = EnemyBullet(pos, vel, name)
        self.add(bullet)

    def update(self, dt):
        for bullet in iter(self):
            bullet.update(dt)
            if bullet.done:
                self.remove(bullet)


class EnemyBullet(pygame.sprite.Sprite):

    def __init__(self, pos, vel, name):
        pygame.sprite.Sprite.__init__(self)

        self.pos = list(pos)
        self.vel = list(vel)

        self.image = pygame.image.load(data.filepath(name)).convert()
        self.image.set_colorkey((255, 0, 255))

        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.done = False

    def update(self, dt):
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        self.rect.center = self.pos
        if util.out_of_bounds(self):
            self.done = True
